﻿using System;
using System.Collections.Generic;
using System.IO;
using CommandLine;
using PaintShopSchedulingSimpleTypes;
using SimpleSolutionValidator.SolutionEvaluator;

namespace SimpleSolutionValidator
{
    public class Start
    {
        public static void Main(string[] args)
        {
            // Parse and process command line arguments
            Parser.Default.ParseArguments<Options>(args)
                .WithParsed(RunValidation)
                .WithNotParsed(HandleParseError);
        }

        private static void PrintUsage()
        {
            Console.WriteLine("Usage: SolutionValidatorPSSP -i <instance-file> -s <solution-file>");
        }

        private static void HandleParseError(IEnumerable<Error> errs)
        {
            PrintUsage();
        }

        private static void RunValidation(Options opts)
        {
            try {
                if (string.IsNullOrEmpty(opts.InstanceFile) || string.IsNullOrEmpty(opts.SolutionFile)) {
                    PrintUsage();
                    return;
                }
                ISimpleInstanceSerializer instanceSerializer =
                    SimpleInstanceSerializerFactory.Instance.GetDefaultSimpleInstanceSerializer();
                SimpleInstance instance = instanceSerializer.DeserializeInstanceFromTxtFile(opts.InstanceFile);
                ISimpleSolution solution =
                    instanceSerializer.DeserializeSolutionFromTextFile(instance, opts.SolutionFile);
                ISimpleSolutionEvaluator solutionEvaluator = new SimpleSolutionEvaluator();

                if (!opts.PrintSingleLineCostOutput) {
                    solutionEvaluator.EvaluateSolution(instance, solution, Console.Out);
                }
                else {
                    string instanceName = Path.GetFileNameWithoutExtension(opts.InstanceFile);
                    solutionEvaluator.EvaluateSolutionSingleLineCost(instanceName, instance, solution, Console.Out);
                }

                if (opts.PrintSolution) {
                    SimpleSolutionTextWriter.WriteSolutionToConsole(instance, solution);
                }

                if (opts.PrintCarriersColorsConfigs) {
                    Console.WriteLine("+++++++++++++++++ Carriers ++++++++++++++++");
                    int maxRound = 0;

                    for (int i = 0; i < solution.NumberOfRounds; i++) {
                        maxRound = Math.Max(maxRound, solution.GetNumberOfCarriersInRound(i));
                    }

                    for (int j = 0; j < maxRound; j++) {
                        for (int i = 0; i < solution.NumberOfRounds; i++) {
                            int carrierTypeId = solution.GetCarrierTypeIdAtPosition(new Position(i, j));
                            if (carrierTypeId > -1) {
                                Console.Write(carrierTypeId.ToString("D2") + " ");
                                //Console.Write((char)(32+carrierTypeId) + " ");
                            }
                            else {
                                Console.Write("   ");
                                //Console.Write(" ");
                            }
                        }

                        Console.WriteLine();
                    }

                    Console.WriteLine();
                    for (int i = 0; i < solution.NumberOfRounds; i++) {
                        int numCarrierBlocks = 0;
                        int previousBlockId = -1;

                        for (int j = 0; j < solution.GetNumberOfCarriersInRound(i); j++) {
                            int carrierTypeId = solution.GetCarrierTypeIdAtPosition(new Position(i, j));
                            if (carrierTypeId != previousBlockId) {
                                numCarrierBlocks++;
                                previousBlockId = carrierTypeId;
                            }
                        }

                        Console.Write(numCarrierBlocks.ToString("D2") + " ");
                    }

                    Console.WriteLine();

                    Console.WriteLine("+++++++++++++++++ Colors ++++++++++++++++");
                    for (int i = 0; i < solution.NumberOfRounds; i++) {
                        maxRound = Math.Max(maxRound, solution.GetNumberOfCarriersInRound(i));
                    }

                    for (int j = 0; j < maxRound; j++) {
                        for (int i = 0; i < solution.NumberOfRounds; i++) {
                            int colorId = solution.GetColorIdAtPosition(new Position(i, j));
                            if (colorId > -1) {
                                Console.Write(colorId.ToString("D2") + " ");
                                //Console.Write((char)(32+carrierTypeId) + " ");
                            }
                            else {
                                Console.Write("   ");
                                //Console.Write(" ");
                            }
                        }

                        Console.WriteLine();
                    }

                    Console.WriteLine();
                    for (int i = 0; i < solution.NumberOfRounds; i++) {
                        int numColorBlocks = 0;
                        int previousBlockId = -1;

                        for (int j = 0; j < solution.GetNumberOfCarriersInRound(i); j++) {
                            int colorBlockId = solution.GetColorIdAtPosition(new Position(i, j));
                            if (colorBlockId != previousBlockId) {
                                numColorBlocks++;
                                previousBlockId = colorBlockId;
                            }
                        }

                        Console.Write(numColorBlocks.ToString("D2") + " ");
                    }

                    Console.WriteLine();
                    Console.WriteLine("+++++++++++++++++ Configs ++++++++++++++++");
                    for (int i = 0; i < solution.NumberOfRounds; i++) {
                        maxRound = Math.Max(maxRound, solution.GetNumberOfCarriersInRound(i));
                    }

                    for (int j = 0; j < maxRound; j++) {
                        for (int i = 0; i < solution.NumberOfRounds; i++) {
                            int configId = solution.GetCarrierConfigurationIdAtPosition(new Position(i, j));
                            if (configId > -1) {
                                Console.Write(configId.ToString("D3") + " ");
                                //Console.Write((char)(32+carrierTypeId) + " ");
                            }
                            else {
                                Console.Write("   ");
                                //Console.Write(" ");
                            }
                        }

                        Console.WriteLine();
                    }

                    Console.WriteLine();
                    for (int i = 0; i < solution.NumberOfRounds; i++) {
                        int numConfigBlocks = 0;
                        int previousBlockId = -1;

                        for (int j = 0; j < solution.GetNumberOfCarriersInRound(i); j++) {
                            int configBlockId = solution.GetCarrierConfigurationIdAtPosition(new Position(i, j));
                            if (configBlockId != previousBlockId) {
                                numConfigBlocks++;
                                previousBlockId = configBlockId;
                            }
                        }

                        Console.Write(numConfigBlocks.ToString("D2") + " ");
                    }

                    Console.WriteLine();
                    Console.WriteLine("+++++++++++++++++ Num Carrier Color Blocks ++++++++++++++++");
                    for (int i = 0; i < solution.NumberOfRounds; i++) {
                        int numCarrierColorBlocks = 0;
                        int previousColorId = -1;
                        int previousCarrierId = -1;

                        for (int j = 0; j < solution.GetNumberOfCarriersInRound(i); j++) {
                            int colorBlockId = solution.GetColorIdAtPosition(new Position(i, j));
                            int carrierBlockId = solution.GetCarrierTypeIdAtPosition(new Position(i, j));
                            if (colorBlockId != previousColorId || carrierBlockId != previousCarrierId) {
                                numCarrierColorBlocks++;
                                previousColorId = colorBlockId;
                                previousCarrierId = carrierBlockId;
                            }
                        }

                        Console.Write(numCarrierColorBlocks.ToString("D2") + " ");
                    }

                    Console.WriteLine();
                }
            }
            catch (Exception) {
                PrintUsage();
            }
        }
    }
}