﻿using System.Collections.Generic;
using PaintShopSchedulingSimpleTypes.SimpleTypes;

namespace PaintShopSchedulingSimpleTypes
{
    /// <summary>
    /// Solution object that will represent a candidate solution to the paint shop scheduling problem
    /// using the simple data type objects.
    /// </summary>
    public interface ISimpleSolution
    {
        /// <summary>
        /// Returns the number of rounds in the schedule
        /// </summary>
        int NumberOfRounds { get; }

        /// <summary>
        /// Return the id of the color that is scheduled at the given position in this solution. 
        /// </summary>
        /// <param name="position">The position in the schedule.</param>
        /// <returns>The id of the color, or -1 if no color is assigned at the given position.</returns>
        int GetColorIdAtPosition(Position position);

        /// <summary>
        /// Return the id of the carrier configuration that is scheduled at the given position in this solution.
        /// </summary>
        /// <param name="position">The position in the schedule.</param>
        /// <returns>The id of the carrier configuration, or -1 if no carrier is scheduled at the given position.</returns>
        int GetCarrierConfigurationIdAtPosition(Position position);

        /// <summary>
        /// Return the id of the carrier type that is scheduled at the given position in this solution.
        /// </summary>
        /// <param name="position">The position in the schedule.</param>
        /// <returns>The id of the carrier type, or -1 if no carrier is scheduled at the given position.</returns>
        int GetCarrierTypeIdAtPosition(Position position);

        /// <summary>
        /// Get information about the carrier that is scheduled at the given position. 
        /// </summary>
        /// <param name="position">The position in the schedule.</param>
        /// <returns>Object with information about the scheduled carrier and assigned color.
        /// The information will be set to -1, if no carrier is assigned at the position.</returns>
        SimpleCarrier GetCarrierAtPosition(Position position);

        /// <summary>
        /// Assign the carrier and color that should be scheduled at the given position. 
        /// WARNING: A simple solution will always assume that all rounds are consistently scheduled (i.e. no unscheduled positions may appear before the last scheduled position within a round),
        /// however the setters will not prevent inconsistent assignments for performance reasons.
        /// Therefore it lies in the responsibility of the caller to keep the solution consistent.
        /// </summary>
        /// <param name="position">The position in the schedule.</param>
        /// <param name="carrier">Object with information about the carrier and color that should be assigned.</param>
        void SetCarrierAtPosition(Position position, SimpleCarrier carrier);

        /// <summary>
        /// Returns the number of currently scheduled carriers in a round.
        /// </summary>
        /// <param name="roundIndex">Index of the target round</param>
        /// <returns>Number of carriers scheduled in the round with the given round index</returns>
        int GetNumberOfCarriersInRound(int roundIndex);

        /// <summary>
        /// Gets all the different carrier type ids that appear in a round.
        /// </summary>
        /// <param name="roundIndex">Index of the target round.</param>
        /// <returns>Collection containing all carrier type ids that appear in the round.</returns>
        IEnumerable<int> GetDistinctCarrierTypeIdsInRound(int roundIndex);

        /// <summary>
        /// Get a list of carrier type ids that are scheduled in a round.
        /// </summary>
        /// <param name="roundIndex">Index of the target round.</param>
        /// <returns>List of all carrier type ids in order as they are scheduled in the target round.</returns>
        List<int> GetCarrierTypeIdsInRound(int roundIndex);
    }
}
