﻿using System;
using System.Collections.Generic;
using System.Linq;
using PaintShopSchedulingSimpleTypes.SimpleTypes;

namespace PaintShopSchedulingSimpleTypes
{
    /// <summary>
    /// WARNING: This serializer might throw any exceptions if inconsistent instances/solutions are processed
    /// Furthermore, carrier/configuration/color-names should not contain the characters [](),;| since they are used
    /// as separators. If an instance uses these characters they will simply be replaced by _ if they are serialized.
    /// However, this can easily lead to inconsisent instance/solution files.
    /// </summary>
    public class SimpleInstanceStringSerializer : ISimpleInstanceStringSerializer
    {
        public ISimpleSolution DeserializeSolutionFromString(SimpleInstance instance, string serializedSolution)
        {
            ISimpleSolution solution =
                new SimpleSolution(instance.NumberOfRounds + 1, instance.MaxCapacity, instance.Carriers.Count - 1);

            string[] serializedLines = serializedSolution.Split(new[] { Environment.NewLine }, StringSplitOptions.None);

            InsertPreviousRound(instance, solution);

            DeserializeSchedule(serializedLines, instance, solution);

            return solution;
        }

        private void InsertPreviousRound(SimpleInstance instance, ISimpleSolution solution)
        {
            for (int position = 0; position < instance.HistoryRound.Count; position++)
            {
                int carrierTypeId = instance.HistoryRound[position].CarrierType.Id;
                int colorId = instance.HistoryRound[position].Color.Id;

                solution.SetCarrierAtPosition(new Position(0, position), new SimpleCarrier(carrierTypeId, -1, colorId));
            }
        }

        private void DeserializeSchedule(string[] serializedLines, SimpleInstance instance, ISimpleSolution solution)
        {
            for (int i = 0; i < serializedLines.Length; i++)
            {
                string serializedLine = serializedLines[i].Trim();

                if ("SCHEDULE" == serializedLine)
                {
                    i++;
                    serializedLine = serializedLines[i].Trim();

                    while (!string.IsNullOrEmpty(serializedLine))
                    {
                        DeserializeSchedulePosition(serializedLine, instance, solution);

                        i++;
                        serializedLine = serializedLines[i].Trim();
                    }
                    break;
                }
            }
        }

        private void DeserializeSchedulePosition(string serializedLine, SimpleInstance instance, ISimpleSolution solution)
        {
            string[] values = serializedLine.Trim('(', ')').Split(',');

            // round index is okay because previous round is 0
            int roundIndex = int.Parse(values[0]);
            // position index has to be corrected because of 0-index
            int posIndex = int.Parse(values[1]) - 1;
            int configId = GetConfigurationByName(values[2], instance).Id;
            SimpleCarrierType carrierType = instance.Carriers.FirstOrDefault(x => x.Configurations.Any(y => y.Id == configId));
            int carrierTypeId = carrierType?.Id ?? -1;
            int colorId = GetColorByName(values[3], instance).Id;

            solution.SetCarrierAtPosition(new Position(roundIndex, posIndex),
                new SimpleCarrier(carrierTypeId, configId, colorId));
        }

        public SimpleInstance DeserializeInstanceFromString(string serializedInstance)
        {
            SimpleInstance instance = new SimpleInstance();

            string[] serializedLines = serializedInstance.Split(new[] { Environment.NewLine }, StringSplitOptions.None);

            DeserializeNumberOfRounds(serializedLines, instance);
            DeserializeRoundCapacity(serializedLines, instance);
            DeserializeColors(serializedLines, instance);
            DeserializeMaterials(serializedLines, instance);
            DeserializeDemands(serializedLines, instance);
            DeserializeConfigurations(serializedLines, instance);
            DeserializeCarrierTypes(serializedLines, instance);
            DeserializeHistoryRound(serializedLines, instance);
            DeserializeAvailableCarriers(serializedLines, instance);
            DeserializeForbiddenCarrierSequences(serializedLines, instance);
            DeserializeMinBlockSizes(serializedLines, instance);
            DeserializeMaxBlockSizes(serializedLines, instance);
            DeserializeForbiddenColorSequences(serializedLines, instance);
            DeserializeColorTransitionCosts(serializedLines, instance);

            return instance;
        }

        private void DeserializeNumberOfRounds(string[] serializedLines, SimpleInstance instance)
        {

            for (int i = 0; i < serializedLines.Length; i++)
            {
                string serializedLine = serializedLines[i].Trim();

                if ("ROUNDS" == serializedLine)
                {
                    i++;
                    string numberOfRounds = serializedLines[i].Trim();

                    instance.NumberOfRounds = int.Parse(numberOfRounds);
                    break;
                }
            }
        }

        private void DeserializeRoundCapacity(string[] serializedLines, SimpleInstance instance)
        {

            for (int i = 0; i < serializedLines.Length; i++)
            {
                string serializedLine = serializedLines[i].Trim();

                if ("ROUNDCAPACITY" == serializedLine)
                {
                    i++;
                    string[] roundCapacities = serializedLines[i].Trim().Split(',');

                    instance.MinCapacity = int.Parse(roundCapacities[0]);
                    instance.MaxCapacity = int.Parse(roundCapacities[1]);

                    break;
                }
            }
        }

        private void DeserializeColors(string[] serializedLines, SimpleInstance instance)
        {

            for (int i = 0; i < serializedLines.Length; i++)
            {
                string serializedLine = serializedLines[i].Trim();

                if ("COLORS" == serializedLine)
                {
                    i++;
                    serializedLine = serializedLines[i].Trim();

                    instance.Colors = new List<SimpleColor>();
                    instance.ColorNames = new List<string>();
                    while (!string.IsNullOrEmpty(serializedLine))
                    {
                        DeserializeColor(serializedLine, instance);

                        i++;
                        serializedLine = serializedLines[i].Trim();
                    }
                    break;
                }
            }
        }

        private void DeserializeColor(string serializedLine, SimpleInstance instance)
        {
            string colorName = serializedLine;
            instance.ColorNames.Add(colorName);
            instance.Colors.Add(new SimpleColor(instance.Colors.Count));
        }


        private void DeserializeMaterials(string[] serializedLines, SimpleInstance instance)
        {
            for (int i = 0; i < serializedLines.Length; i++)
            {
                string serializedLine = serializedLines[i].Trim();

                if ("MATERIALS" == serializedLine)
                {
                    i++;
                    serializedLine = serializedLines[i].Trim();

                    instance.Materials = new List<SimpleMaterial>();
                    instance.MaterialNames = new List<string>();
                    while (!string.IsNullOrEmpty(serializedLine))
                    {
                        DeserializeMaterial(serializedLine, instance);

                        i++;
                        serializedLine = serializedLines[i].Trim();
                    }
                    break;
                }
            }
        }

        private void DeserializeMaterial(string serializedLine, SimpleInstance instance)
        {
            string materialName = serializedLine;
            instance.MaterialNames.Add(materialName);
            instance.Materials.Add(new SimpleMaterial(instance.Materials.Count));
        }

        private void DeserializeDemands(string[] serializedLines, SimpleInstance instance)
        {
            for (int i = 0; i < serializedLines.Length; i++)
            {
                string serializedLine = serializedLines[i].Trim();

                if ("DEMANDS" == serializedLine)
                {
                    i++;
                    serializedLine = serializedLines[i].Trim();

                    instance.Demands = new List<SimpleDemand>();
                    while (!string.IsNullOrEmpty(serializedLine))
                    {
                        DeserializeDemand(serializedLine, instance);

                        i++;
                        serializedLine = serializedLines[i].Trim();
                    }
                    break;
                }
            }
        }

        private void DeserializeDemand(string serializedLine, SimpleInstance instance)
        {
            string[] values = serializedLine.Trim('(', ')').Split(',');

            int numberOfPieces = int.Parse(values[0]);
            SimpleMaterial material = GetMaterialByName(values[1], instance);
            int dueRound = int.Parse(values[2]);
            SimpleColor color = GetColorByName(values[3], instance);

            instance.Demands.Add(new SimpleDemand(instance.Demands.Count, numberOfPieces, material, dueRound, color));
        }

        private void DeserializeConfigurations(string[] serializedLines, SimpleInstance instance)
        {

            for (int i = 0; i < serializedLines.Length; i++)
            {
                string serializedLine = serializedLines[i].Trim();

                if ("CONFIGURATIONS" == serializedLine)
                {
                    serializedLine = serializedLines[i + 1].Trim();

                    instance.Configurations = new List<SimpleCarrierConfiguration>();
                    instance.ConfigurationNames = new List<string>();
                    while (!string.IsNullOrEmpty(serializedLine))
                    {
                        DeserializeConfiguration(serializedLine, instance);

                        i++;
                        serializedLine = serializedLines[i + 1].Trim();
                    }
                    break;
                }
            }
        }

        private void DeserializeConfiguration(string serializedLine, SimpleInstance instance)
        {
            string[] values = serializedLine.Split(',');

            instance.ConfigurationNames.Add(values[0]);

            string[] stringMaterialCapacities = values[1].Trim('[', ']').Split('|');

            ICollection<SimpleMaterialCapacity> materialCapacities = new List<SimpleMaterialCapacity>();

            foreach (string materialCapacity in stringMaterialCapacities)
            {
                string[] capacityValues = materialCapacity.Trim('(', ')').Split(';');
                int capacity = int.Parse(capacityValues[0]);
                SimpleMaterial material = GetMaterialByName(capacityValues[1], instance);

                materialCapacities.Add(new SimpleMaterialCapacity(material, capacity));
            }

            instance.Configurations.Add(new SimpleCarrierConfiguration(instance.Configurations.Count, materialCapacities));
        }

        private void DeserializeCarrierTypes(string[] serializedLines, SimpleInstance instance)
        {

            for (int i = 0; i < serializedLines.Length; i++)
            {
                string serializedLine = serializedLines[i].Trim();

                if ("CARRIERS" == serializedLine)
                {
                    i++;
                    serializedLine = serializedLines[i].Trim();

                    instance.Carriers = new List<SimpleCarrierType>();
                    instance.CarrierTypeNames = new List<string>();
                    instance.CarrierTypeInfos = new List<SimpleCarrierTypeInformation>();
                    while (!string.IsNullOrEmpty(serializedLine))
                    {
                        DeserializeCarrierType(serializedLine, instance);

                        i++;
                        serializedLine = serializedLines[i].Trim();
                    }
                    break;
                }
            }
        }

        private void DeserializeCarrierType(string serializedLine, SimpleInstance instance)
        {
            string[] values = serializedLine.Split(',');

            instance.CarrierTypeNames.Add(values[0]);

            string[] stringConfigurations = values[1].Trim('[', ']').Split('|');

            ICollection<SimpleCarrierConfiguration> carrierConfigurations = new List<SimpleCarrierConfiguration>();

            foreach (string carrierConfigurationName in stringConfigurations)
            {
                if (!string.IsNullOrEmpty(carrierConfigurationName))
                {
                    carrierConfigurations.Add(GetConfigurationByName(carrierConfigurationName, instance));
                }
            }

            SimpleCarrierType carrierType = new SimpleCarrierType(instance.Carriers.Count, carrierConfigurations);
            instance.Carriers.Add(carrierType);
            instance.CarrierTypeInfos.Add(new SimpleCarrierTypeInformation(carrierType, 0, 0));
        }

        private void DeserializeHistoryRound(string[] serializedLines, SimpleInstance instance)
        {

            for (int i = 0; i < serializedLines.Length; i++)
            {
                string serializedLine = serializedLines[i].Trim();

                if ("HISTORY_ROUND" == serializedLine)
                {
                    i++;
                    serializedLine = serializedLines[i].Trim();

                    instance.HistoryRound = new List<SimpleHistoryRoundPosition>();
                    while (!string.IsNullOrEmpty(serializedLine))
                    {
                        DeserializeHistoryRoundPosition(serializedLine, instance);
                        i++;
                        serializedLine = serializedLines[i].Trim();
                    }
                    break;
                }
            }
        }

        private void DeserializeHistoryRoundPosition(string serializedLine, SimpleInstance instance)
        {
            string[] values = serializedLine.Trim('(', ')').Split(',');

            int position = int.Parse(values[0]);
            SimpleCarrierType carrierType = GetCarrierTypeByName(values[1], instance);
            SimpleColor color = GetColorByName(values[2], instance);

            if (instance.HistoryRound.Count < position)
            {
                // if position goes out of bound fill the list with null values until positions
                for (int i = instance.HistoryRound.Count; i < position; i++)
                {
                    instance.HistoryRound.Add(null);
                }
            }
            instance.HistoryRound[position - 1] = new SimpleHistoryRoundPosition(carrierType, color);
        }

        private void DeserializeAvailableCarriers(string[] serializedLines, SimpleInstance instance)
        {

            for (int i = 0; i < serializedLines.Length; i++)
            {
                string serializedLine = serializedLines[i].Trim();

                if ("AVAILABLE_CARRIERS" == serializedLine)
                {
                    i++;
                    serializedLine = serializedLines[i].Trim();

                    ICollection<SimpleCarrierTypeInformation> carrierTypeInfos = new List<SimpleCarrierTypeInformation>();
                    while (!string.IsNullOrEmpty(serializedLine))
                    {
                        carrierTypeInfos.Add(DeserializeCarrierAvailabilities(serializedLine, instance));

                        i++;
                        serializedLine = serializedLines[i].Trim();
                    }

                    instance.CarrierTypeInfos = carrierTypeInfos;
                    break;
                }
            }
        }

        private SimpleCarrierTypeInformation DeserializeCarrierAvailabilities(string serializedLine, SimpleInstance instance)
        {
            string[] values = serializedLine.Split(',');

            SimpleCarrierType carrierType = GetCarrierTypeByName(values[0], instance);
            string[] stringAvailabilities = values[1].Trim('[', ']').Split('|');
            IList<int> carrierAvailabilities = stringAvailabilities.Select(int.Parse).ToList();

            SimpleCarrierTypeInformation oldCarrierInfo =
                instance.CarrierTypeInfos.FirstOrDefault(x => x.CarrierType.Id == carrierType.Id);

            if (oldCarrierInfo == null)
            {
                return null;
            }

            return new SimpleCarrierTypeInformation(oldCarrierInfo.CarrierType, oldCarrierInfo.MinimumBlockSize,
                oldCarrierInfo.MaximumBlockSize, carrierAvailabilities);
        }


        private void DeserializeForbiddenCarrierSequences(string[] serializedLines, SimpleInstance instance)
        {

            for (int i = 0; i < serializedLines.Length; i++)
            {
                string serializedLine = serializedLines[i].Trim();

                if ("FORBIDDEN_CARRIER_SEQUENCES" == serializedLine)
                {
                    i++;
                    serializedLine = serializedLines[i].Trim();

                    instance.ForbiddenCarrierSequences = new List<SimpleForbiddenCarrierTypeSequence>();
                    while (!string.IsNullOrEmpty(serializedLine))
                    {
                        instance.ForbiddenCarrierSequences.Add(DeserializeForbiddenCarrierTypeSequence(serializedLine, instance));

                        i++;
                        serializedLine = serializedLines[i].Trim();
                    }
                    break;
                }
            }
        }

        private SimpleForbiddenCarrierTypeSequence DeserializeForbiddenCarrierTypeSequence(string serializedLine, SimpleInstance instance)
        {
            string[] values = serializedLine.Trim('(', ')').Split(',');

            SimpleCarrierType fromCarrier = GetCarrierTypeByName(values[0], instance);
            SimpleCarrierType toCarrier = GetCarrierTypeByName(values[1], instance);

            return new SimpleForbiddenCarrierTypeSequence(fromCarrier, toCarrier);
        }

        private void DeserializeMinBlockSizes(string[] serializedLines, SimpleInstance instance)
        {

            for (int i = 0; i < serializedLines.Length; i++)
            {
                string serializedLine = serializedLines[i].Trim();

                if ("MINBLOCK_SIZES" == serializedLine)
                {
                    i++;
                    serializedLine = serializedLines[i].Trim();

                    ICollection<SimpleCarrierTypeInformation> carrierTypeInformationsWithMinBlock = new List<SimpleCarrierTypeInformation>();
                    while (!string.IsNullOrEmpty(serializedLine))
                    {
                        carrierTypeInformationsWithMinBlock.Add(DeserializeMinBlockSize(serializedLine, instance));

                        i++;
                        serializedLine = serializedLines[i].Trim();
                    }

                    instance.CarrierTypeInfos = carrierTypeInformationsWithMinBlock;
                    break;
                }
            }
        }

        private SimpleCarrierTypeInformation DeserializeMinBlockSize(string serializedLine, SimpleInstance instance)
        {
            string[] values = serializedLine.Trim('(', ')').Split(',');

            SimpleCarrierType carrierType = GetCarrierTypeByName(values[0], instance);
            int minBlockSize = int.Parse(values[1]);
            SimpleCarrierTypeInformation oldCarrierInfo =
                instance.CarrierTypeInfos.FirstOrDefault(x => x.CarrierType.Id == carrierType.Id);

            if (oldCarrierInfo == null)
            {
                return null;
            }

            return new SimpleCarrierTypeInformation(oldCarrierInfo.CarrierType, minBlockSize,
                oldCarrierInfo.MaximumBlockSize, oldCarrierInfo.Availabilities);
        }
        private void DeserializeMaxBlockSizes(string[] serializedLines, SimpleInstance instance)
        {

            for (int i = 0; i < serializedLines.Length; i++)
            {
                string serializedLine = serializedLines[i].Trim();

                if ("MAXBLOCK_SIZES" == serializedLine)
                {
                    i++;
                    serializedLine = serializedLines[i].Trim();

                    ICollection<SimpleCarrierTypeInformation> carrierTypeInformationsWithMaxBlock = new List<SimpleCarrierTypeInformation>();
                    while (!string.IsNullOrEmpty(serializedLine))
                    {
                        carrierTypeInformationsWithMaxBlock.Add(DeserializeMaxBlockSize(serializedLine, instance));
                        i++;
                        serializedLine = serializedLines[i].Trim();
                    }

                    instance.CarrierTypeInfos = carrierTypeInformationsWithMaxBlock;
                    break;
                }
            }
        }

        private SimpleCarrierTypeInformation DeserializeMaxBlockSize(string serializedLine, SimpleInstance instance)
        {
            string[] values = serializedLine.Trim('(', ')').Split(',');

            SimpleCarrierType carrierType = GetCarrierTypeByName(values[0], instance);
            int maxBlockSize = int.Parse(values[1]);
            SimpleCarrierTypeInformation oldCarrierInfo =
                instance.CarrierTypeInfos.FirstOrDefault(x => x.CarrierType.Id == carrierType.Id);

            if (oldCarrierInfo == null)
            {
                return null;
            }

            return new SimpleCarrierTypeInformation(oldCarrierInfo.CarrierType, oldCarrierInfo.MinimumBlockSize,
                maxBlockSize, oldCarrierInfo.Availabilities);
        }

        private void DeserializeForbiddenColorSequences(string[] serializedLines, SimpleInstance instance)
        {

            for (int i = 0; i < serializedLines.Length; i++)
            {
                string serializedLine = serializedLines[i].Trim();

                if ("FORBIDDEN_COLOR_SEQUENCES" == serializedLine)
                {
                    i++;
                    serializedLine = serializedLines[i].Trim();

                    instance.ForbiddenColorSequences = new List<SimpleForbiddenColorSequence>();
                    while (!string.IsNullOrEmpty(serializedLine))
                    {
                        instance.ForbiddenColorSequences.Add(DeserializeForbiddenColorSequence(serializedLine, instance));

                        i++;
                        serializedLine = serializedLines[i].Trim();
                    }
                    break;
                }
            }
        }

        private SimpleForbiddenColorSequence DeserializeForbiddenColorSequence(string serializedLine, SimpleInstance instance)
        {
            string[] values = serializedLine.Trim('(', ')').Split(',');

            SimpleColor fromColor = GetColorByName(values[0], instance);
            SimpleColor toColor = GetColorByName(values[1], instance);
            int minCarriersBetween = int.Parse(values[2]);

            return new SimpleForbiddenColorSequence(fromColor, toColor, minCarriersBetween);
        }

        private void DeserializeColorTransitionCosts(string[] serializedLines, SimpleInstance instance)
        {

            for (int i = 0; i < serializedLines.Length; i++)
            {
                string serializedLine = serializedLines[i].Trim();

                if ("COLOR_TRANSITION_COST" == serializedLine)
                {
                    i++;
                    serializedLine = serializedLines[i].Trim();

                    instance.ColorTransitionCosts = new List<SimpleColorTransitionCost>();
                    while (!string.IsNullOrEmpty(serializedLine))
                    {
                        instance.ColorTransitionCosts.Add(DeserializeColorTransitionCost(serializedLine, instance));

                        i++;
                        serializedLine = serializedLines[i].Trim();
                    }
                    break;
                }
            }
        }

        private SimpleColorTransitionCost DeserializeColorTransitionCost(string serializedLine, SimpleInstance instance)
        {
            string[] values = serializedLine.Trim('(', ')').Split(',');

            SimpleColor fromColor = GetColorByName(values[0], instance);
            SimpleColor toColor = GetColorByName(values[1], instance);
            int cost = int.Parse(values[2]);

            return new SimpleColorTransitionCost(fromColor, toColor, cost);
        }

        private SimpleColor GetColorByName(string name, SimpleInstance instance)
        {
            if (instance.Colors.Any(x => instance.ColorNames[x.Id] == name))
            {
                return instance.Colors.First(x => instance.ColorNames[x.Id] == name);
            }
            else
            {
                return new SimpleColor(-1);
            }
        }

        private SimpleMaterial GetMaterialByName(string name, SimpleInstance instance)
        {
            return instance.Materials.First(x => instance.MaterialNames[x.Id] == name);
        }

        private SimpleCarrierType GetCarrierTypeByName(string name, SimpleInstance instance)
        {
            return instance.Carriers.First(x => instance.CarrierTypeNames[x.Id] == name);
        }

        private SimpleCarrierConfiguration GetConfigurationByName(string name, SimpleInstance instance)
        {
            return instance.Configurations.First(x => instance.ConfigurationNames[x.Id] == name);
        }
    }
}
