﻿using System;
using System.Collections.Generic;
using System.Linq;
using PaintShopSchedulingSimpleTypes.SimpleTypes;

namespace PaintShopSchedulingSimpleTypes
{
    public class SimpleSolution : ISimpleSolution
    {

        private readonly SimpleCarrier[] _assignedCarriers;
        private readonly int[] _assignedCarrierTypeIds;
        private readonly int[] _assignedColorIds;
        private readonly int[] _assignedConfigurationIds;
        private readonly int[] _numberOfCarriersPerRound;
        private readonly ICollection<int>[] _carrierTypesAppearingInRound;
        private readonly List<int>[] _numberOfCarriersWithTypePerRound;

        public int NumberOfRounds { get; }
        public int MaxRoundCapacity { get; }
      
        /// <summary>
        /// Create an empty solution.
        /// </summary>
        /// <param name="numberOfRounds">Number of rounds including the history round.</param>
        /// <param name="maxCarriersPerRound">Max carrier capacity per round</param>
        /// <param name="maxCarrierTypeId">Maximum carrier type id.</param>
        public SimpleSolution(int numberOfRounds, int maxCarriersPerRound, int maxCarrierTypeId)
        {
            NumberOfRounds = numberOfRounds;
            MaxRoundCapacity = maxCarriersPerRound;

            // assign all positions with -1 empty

            int maxTotalNumberOfCarriers = NumberOfRounds * MaxRoundCapacity;
            _assignedCarriers = new SimpleCarrier[maxTotalNumberOfCarriers];
            _assignedCarrierTypeIds = new int[maxTotalNumberOfCarriers];
            _assignedColorIds = new int[maxTotalNumberOfCarriers];
            _assignedConfigurationIds = new int[maxTotalNumberOfCarriers];
            for (int i = 0; i < maxTotalNumberOfCarriers; i++)
            {
                _assignedCarriers[i] = new SimpleCarrier(-1, -1, -1);
                _assignedCarrierTypeIds[i] = -1;
                _assignedColorIds[i] = -1;
                _assignedConfigurationIds[i] = -1;
            }

            // initialize number of carriers per round and carrier types appearing in each round
            _numberOfCarriersPerRound = new int[numberOfRounds];
            _carrierTypesAppearingInRound = new ICollection<int>[numberOfRounds];
            _numberOfCarriersWithTypePerRound = new List<int>[numberOfRounds];
            for (int i = 0; i < numberOfRounds; i++)
            {
                _numberOfCarriersPerRound[i] = 0;
                _carrierTypesAppearingInRound[i] = new List<int>();
                _numberOfCarriersWithTypePerRound[i] = Enumerable.Repeat(0, maxCarrierTypeId + 1).ToList();
            }

        }

        public int GetColorIdAtPosition(Position position)
        {
            return _assignedColorIds[position.RoundIndex * MaxRoundCapacity + position.PositionIndex];
        }

        public int GetCarrierConfigurationIdAtPosition(Position position)
        {
            return _assignedConfigurationIds[position.RoundIndex * MaxRoundCapacity + position.PositionIndex];
        }

        public int GetCarrierTypeIdAtPosition(Position position)
        {
            return _assignedCarrierTypeIds[position.RoundIndex * MaxRoundCapacity + position.PositionIndex];
        }

        public void SetColorIdAtPosition(Position position, int colorId)
        {
            int roundIndex = position.RoundIndex;
            int positionIndex = position.PositionIndex;
            _assignedColorIds[roundIndex * MaxRoundCapacity + positionIndex] = colorId;
            SimpleCarrier simpleCarrier = _assignedCarriers[roundIndex * MaxRoundCapacity + positionIndex];
            simpleCarrier.ColorId = colorId;
            _assignedCarriers[roundIndex * MaxRoundCapacity + positionIndex] = simpleCarrier;
        }

        public void SetCarrierConfigurationIdAtPosition(Position position, int carrierConfigurationId)
        {
            int roundIndex = position.RoundIndex;
            int positionIndex = position.PositionIndex;
            _assignedConfigurationIds[roundIndex * MaxRoundCapacity + positionIndex] = carrierConfigurationId;
            SimpleCarrier simpleCarrier = _assignedCarriers[roundIndex * MaxRoundCapacity + positionIndex];
            simpleCarrier.CarrierConfigurationId = carrierConfigurationId;
            _assignedCarriers[roundIndex * MaxRoundCapacity + positionIndex] = simpleCarrier;
        }

        public void SetCarrierTypeIdAtPosition(Position position, int newCarrierTypeId)
        {
            int roundIndex = position.RoundIndex;
            int positionIndex = position.PositionIndex;
            int oldCarrierTypeId = _assignedCarrierTypeIds[roundIndex * MaxRoundCapacity + positionIndex];

            if (oldCarrierTypeId != newCarrierTypeId)
            {
                if (oldCarrierTypeId == -1)
                {
                    _numberOfCarriersPerRound[roundIndex]++;
                }
                else
                {
                    if (newCarrierTypeId == -1)
                    {
                        _numberOfCarriersPerRound[roundIndex]--;
                    }
                    _numberOfCarriersWithTypePerRound[roundIndex][oldCarrierTypeId]--;
                    if (_numberOfCarriersWithTypePerRound[roundIndex][oldCarrierTypeId] == 0)
                    {
                        _carrierTypesAppearingInRound[roundIndex].Remove(oldCarrierTypeId);
                    }
                }

                if (newCarrierTypeId > -1)
                {
                    _numberOfCarriersWithTypePerRound[roundIndex][newCarrierTypeId]++;
                    if (_numberOfCarriersWithTypePerRound[roundIndex][newCarrierTypeId] == 1)
                    {
                        _carrierTypesAppearingInRound[roundIndex].Add(newCarrierTypeId);
                    }
                }
            }
            //
            // set new assigned carrier type ids
            //
            _assignedCarrierTypeIds[roundIndex * MaxRoundCapacity + positionIndex] = newCarrierTypeId;

            SimpleCarrier simpleCarrier = _assignedCarriers[roundIndex * MaxRoundCapacity + positionIndex];
            simpleCarrier.CarrierTypeId = newCarrierTypeId;
            _assignedCarriers[roundIndex * MaxRoundCapacity + positionIndex] = simpleCarrier;
        }

        public SimpleCarrier GetCarrierAtPosition(Position position)
        {
            int roundIndex = position.RoundIndex;
            int positionIndex = position.PositionIndex;
            return _assignedCarriers[roundIndex * MaxRoundCapacity + positionIndex];
        }

        public void SetCarrierAtPosition(Position position, SimpleCarrier carrier)
        {
            SetColorIdAtPosition(position, carrier.ColorId);
            SetCarrierConfigurationIdAtPosition(position, carrier.CarrierConfigurationId);
            SetCarrierTypeIdAtPosition(position, carrier.CarrierTypeId);
        }

        public int GetNumberOfCarriersInRound(int roundIndex)
        {
            return _numberOfCarriersPerRound[roundIndex];
        }

        public IEnumerable<int> GetDistinctCarrierTypeIdsInRound(int roundIndex)
        {
            return _carrierTypesAppearingInRound[roundIndex];
        }

        public List<int> GetCarrierTypeIdsInRound(int roundIndex)
        {
            int numCarriers = _numberOfCarriersPerRound[roundIndex];

            int[] carrierTypeIdsInRound = new int[numCarriers];
            Array.Copy(_assignedCarrierTypeIds, roundIndex * MaxRoundCapacity, carrierTypeIdsInRound, 0, numCarriers);
            return new List<int>(carrierTypeIdsInRound);
        }
    }
}
