﻿using System;
using System.Collections.Generic;
using System.IO;
using PaintShopSchedulingSimpleTypes.SimpleTypes;
using PaintShopSchedulingCarrierChangeOptimizer;

namespace PaintShopSchedulingSimpleTypes
{
    /// <summary>
    /// Provides methods to write text based solution visualizations
    /// </summary>
    public static class SimpleSolutionTextWriter
    {
        /// <summary>
        /// The text description length can be customized
        /// </summary>
        private const int TEXT_DESC_LENGTH = 6;

        /// <summary>
        /// The id description length can be customized
        /// </summary>
        private const int ID_DESC_LENGTH = 3;

        /// <summary>
        /// The cell length is calculated from the given text/id description length
        /// </summary>
        private const int CELL_LENGTH = TEXT_DESC_LENGTH + ID_DESC_LENGTH + 6;

        /// <summary>
        /// A set of conflicted cells can be injected.
        /// The conflicted cell coordinates will then be marked with a !! sign in the output
        /// </summary>
        public static ISet<Position> Conflicts = new HashSet<Position>();

        /// <summary>
        /// Solution that will be visualized
        /// </summary>
        private static ISimpleSolution _solution;

        /// <summary>
        /// Information about the insertion and removals in the solution  
        /// </summary>
        private static IList<IRoundInformation> _insertionRemovalFlagsForRounds;

        /// <summary>
        /// Instance information about the solution
        /// </summary>
        private static SimpleInstance _instance;

        /// <summary>
        /// Write a visualization of the given solution to console output.
        /// </summary>
        /// <param name="instance">Instance information about the solution that should be visualized.</param>
        /// <param name="solution">The solution that should be visualized.</param>
        public static void WriteSolutionToConsole(SimpleInstance instance, ISimpleSolution solution)
        {
            WriteSolution(instance, solution, Console.Out);
        }

        /// <summary>
        /// Write a visualization of the given solution using the given textwriter.
        /// </summary>
        /// <param name="instance">Instance information about the solution that should be visualized.</param>
        /// <param name="solution">The solution that should be visualized.</param>
        /// <param name="textWriter">The textWriter that will be used to write the solution.</param>
        public static void WriteSolution(SimpleInstance instance, ISimpleSolution solution, TextWriter textWriter)
        {
            _instance = instance;
            _solution = solution;
            CalculateInsertionRemovalFlags();

            WriteHeader(_solution.NumberOfRounds, textWriter);
            WriteRows(textWriter);
        }

        private static void WriteRows(TextWriter textWriter)
        {
            int maxRowNumber = CalculateMaximumRowNumber();
            for (int rowIndex = 0; rowIndex < maxRowNumber; rowIndex++)
            {
                WriteRowLine(rowIndex, textWriter, WritePositionCell);
                WriteRowLine(rowIndex, textWriter, WriteCarrierTypeCell);
                WriteRowLine(rowIndex, textWriter, WriteConfigCell);
                WriteRowLine(rowIndex, textWriter, WriteColorCell);

                WriteHorizontalLine(_solution.NumberOfRounds, '-', textWriter);
            }
        }

        private static void WritePositionCell(SimpleCarrier carrierInSchedule, Position position, TextWriter textWriter)
        {
            textWriter.Write("  ");
            textWriter.Write(position.RoundIndex.ToString("D3"));
            textWriter.Write("-");
            textWriter.Write(position.PositionIndex.ToString("D3"));
            textWriter.Write(new string(' ', CELL_LENGTH - 11));
            textWriter.Write("  |");
        }

        private static void WriteRowLine(int rowIndex, TextWriter textWriter, Action<SimpleCarrier, Position, TextWriter> carrierCellWriter)
        {
            textWriter.Write("|");
            for (int colIndex = 0; colIndex < _solution.NumberOfRounds; colIndex++)
            {
                if (_solution.GetNumberOfCarriersInRound(colIndex) > rowIndex)
                {
                    SimpleCarrier carrier = _solution.GetCarrierAtPosition(new Position(colIndex, rowIndex));
                    carrierCellWriter(carrier, new Position(colIndex, rowIndex), textWriter);
                }
                else
                {
                    WriteEmptyCell(textWriter);
                }
            }

            textWriter.WriteLine();
        }

        private static void WriteCarrierTypeCell(SimpleCarrier carrierInSchedule, Position position, TextWriter textWriter)
        {
            textWriter.Write("  ");
            textWriter.Write(FormatTextDescription(GetCarrierTypeName(carrierInSchedule.CarrierTypeId)));
            textWriter.Write("  ");
            textWriter.Write(FormatIdDescription(carrierInSchedule.CarrierTypeId));
            textWriter.Write("  |");
        }

        private static void WriteConfigCell(SimpleCarrier carrierInSchedule, Position position, TextWriter textWriter)
        {
            textWriter.Write(GetInsertionForPosition(position.RoundIndex, position.PositionIndex) ? "+" : " ");
            textWriter.Write(" ");

            textWriter.Write(FormatTextDescription(GetConfigurationName(carrierInSchedule.CarrierConfigurationId)));

            textWriter.Write(Conflicts.Contains(position) ? "!!" : "  ");

            textWriter.Write(FormatIdDescription(carrierInSchedule.CarrierConfigurationId));

            textWriter.Write(" ");
            textWriter.Write(GetRemovalForPosition(position.RoundIndex, position.PositionIndex) ? "-" : " ");
            textWriter.Write("|");
        }

        private static void WriteColorCell(SimpleCarrier carrierInSchedule, Position position, TextWriter textWriter)
        {
            textWriter.Write("  ");
            textWriter.Write(FormatTextDescription(GetColorName(carrierInSchedule.ColorId)));
            textWriter.Write(HasColorChangedToPreviousPosition(carrierInSchedule, position) ? "*" : " ");
            textWriter.Write(" ");
            textWriter.Write(FormatIdDescription(carrierInSchedule.ColorId));
            textWriter.Write("  |");
        }

        private static bool HasColorChangedToPreviousPosition(SimpleCarrier carrier, Position position)
        {
            Position previousPosition = PositionCalculator.GetPreviousPosition(position, _solution);

            if (previousPosition.RoundIndex == -1)
            {
                return false;
            }

            SimpleCarrier previousCarrier = _solution.GetCarrierAtPosition(previousPosition);

            return carrier.ColorId != previousCarrier.ColorId;
        }

        private static void WriteEmptyCell(TextWriter textWriter)
        {
            textWriter.Write(new string(' ', CELL_LENGTH) + "|");
        }

        private static string FormatIdDescription(int typeId)
        {
            string idDescriptionFormat = "{0,-" + ID_DESC_LENGTH + "}";
            string idDescription = string.Format(idDescriptionFormat, typeId);
            return idDescription.Substring(0, ID_DESC_LENGTH);
        }

        private static string FormatTextDescription(string typeName)
        {
            string textDescriptionFormat = "{0,-" + TEXT_DESC_LENGTH + "}";
            string textDescription = string.Format(textDescriptionFormat, typeName);
            return textDescription.Substring(0, TEXT_DESC_LENGTH);
        }

        private static void WriteHeader(int numberOfRounds, TextWriter textWriter)
        {
            WriteHorizontalLine(numberOfRounds, '=', textWriter);

            textWriter.Write("|");

            for (int roundIndex = 0; roundIndex <= numberOfRounds - 1; roundIndex++)
            {
                textWriter.Write(" Round ");

                string cellLengthFormat = "{0,-" + (CELL_LENGTH - 7) + "}";
                string numString = string.Format(cellLengthFormat, roundIndex);

                textWriter.Write(numString.Substring(0, CELL_LENGTH - 7));

                textWriter.Write("|");
            }

            textWriter.WriteLine();

            WriteHorizontalLine(numberOfRounds, '=', textWriter);
        }

        private static int CalculateMaximumRowNumber()
        {
            int maxRowNumber = 0;

            for (int roundIndex = 0; roundIndex < _solution.NumberOfRounds; roundIndex++)
            {
                maxRowNumber = Math.Max(maxRowNumber, _solution.GetNumberOfCarriersInRound(roundIndex));
            }

            return maxRowNumber;
        }

        private static void WriteHorizontalLine(int numberOfRounds, char lineChar, TextWriter textWriter)
        {
            textWriter.WriteLine(new string(lineChar, (CELL_LENGTH + 1) * numberOfRounds + 1));
        }


        private static bool GetInsertionForPosition(int roundIndex, int positionIndex)
        {
            if (_insertionRemovalFlagsForRounds.Count <= roundIndex)
                return false;

            if (_insertionRemovalFlagsForRounds[roundIndex].CarrierInsertionRemovalFlags.Count <= positionIndex) {
                return false;
            }

            return _insertionRemovalFlagsForRounds[roundIndex].CarrierInsertionRemovalFlags[positionIndex].InsertionFlag;
        }

        private static bool GetRemovalForPosition(int roundIndex, int positionIndex)
        {
            if (_insertionRemovalFlagsForRounds.Count <= roundIndex)
                return false;

            if (_insertionRemovalFlagsForRounds[roundIndex].CarrierInsertionRemovalFlags.Count <= positionIndex) {
                return false;
            }

            return _insertionRemovalFlagsForRounds[roundIndex].CarrierInsertionRemovalFlags[positionIndex].RemovalFlag;
        }

        private static string GetCarrierTypeName(int carrierTypeId)
        {
            if (carrierTypeId >= 0 && carrierTypeId < _instance.CarrierTypeNames.Count) {
                return _instance.CarrierTypeNames[carrierTypeId];
            }

            return "NA";
        }

        private static string GetConfigurationName(int configurationId)
        {
            if (configurationId >= 0 && configurationId < _instance.ConfigurationNames.Count) {
                return _instance.ConfigurationNames[configurationId];
            }

            return "NA";
        }

        private static string GetColorName(int colorId)
        {
            if (colorId >= 0 && colorId < _instance.ColorNames.Count) {
                return _instance.ColorNames[colorId];
            }

            return "NA";
        }

        /// <summary>
        /// Calculate insertion and removals for the solution using the default carrier change optimization algorithm
        /// </summary>
        private static void CalculateInsertionRemovalFlags()
        {
            _insertionRemovalFlagsForRounds = new List<IRoundInformation>();
            for (int roundIndex = 0; roundIndex < _solution.NumberOfRounds; roundIndex++)
            {
                _insertionRemovalFlagsForRounds.Add(new SimpleRoundInformation(_solution.GetCarrierTypeIdsInRound(roundIndex)));
            }

            ICarrierChangeOptimizer carrierChangeOptimizer =
                CarrierChangeOptimizerFactory.Instance.GetDefaultCarrierChangeOptimizer();

            for (int roundIndex = 0; roundIndex < _solution.NumberOfRounds-1; roundIndex++)
            {
                carrierChangeOptimizer.Solve(_insertionRemovalFlagsForRounds[roundIndex],
                    _insertionRemovalFlagsForRounds[roundIndex + 1]);
            }
        }

    }
}
