﻿using System;
using System.Collections.Generic;
using System.Linq;
using PaintShopSchedulingSimpleTypes;
using PaintShopSchedulingCarrierChangeOptimizer;

namespace SimpleSolutionValidator.SolutionEvaluator
{
    public class CarrierChangeEvaluator : ICarrierChangeEvaluator
    {
        private int _totalChanges;
        // changes per round = insertions & removals for each round
        private IList<int> _changesPerRound;
        // changes after round = removals after this round & insertions into next round
        private IList<int> _changesAfterRound;
        private readonly ICarrierChangeOptimizer _carrierChangeOptimizer;

        public CarrierChangeEvaluator()
        {
            _carrierChangeOptimizer =
                CarrierChangeOptimizerFactory.Instance.GetSimpleLongestCommonSubsequenceCarrierChangeOptimizer();
        }

        public void Evaluate(ISimpleSolution solution)
        {
            _totalChanges = 0;
            _changesPerRound = Enumerable.Repeat(0, solution.NumberOfRounds).ToList();
            _changesAfterRound = Enumerable.Repeat(0, solution.NumberOfRounds-1).ToList();

            for (int roundIndex = 0; roundIndex < solution.NumberOfRounds-1; roundIndex++) {
                IList<int> carrierTypeIdsCurrentRound = solution.GetCarrierTypeIdsInRound(roundIndex);
                IList<int> carrierTypeIdsNextRound = solution.GetCarrierTypeIdsInRound(roundIndex+1);
                IRoundInformation roundInformationCurrentRound = new SimpleRoundInformation(carrierTypeIdsCurrentRound);
                IRoundInformation roundInformationNextRound = new SimpleRoundInformation(carrierTypeIdsNextRound);

                INumberOfCarrierChanges carrierChanges = _carrierChangeOptimizer.EvaluateNumberOfChanges(roundInformationCurrentRound, roundInformationNextRound);
                int changesCurrentRound = carrierChanges.NumberOfCarrierRemovalsOfCurrentRound;
                int changesNextRound = carrierChanges.NumberOfCarrierInsertionsOfNextRound;


                _totalChanges += changesCurrentRound + changesNextRound;

                _changesPerRound[roundIndex] += changesCurrentRound;
                _changesPerRound[roundIndex+1] += changesNextRound;
                _changesAfterRound[roundIndex] += changesCurrentRound + changesNextRound;
            }
        }

        public int GetTotalNumberOfChanges()
        {
            return _totalChanges;
        }
        public int GetSquaredNumberOfChanges()
        {
            int totalSquaredChanges = 0;
            foreach (int changesAfterRound in _changesAfterRound) {
                totalSquaredChanges += changesAfterRound * changesAfterRound;
            }

            return totalSquaredChanges;
        }
        
        public IList<int> GetCarrierChangesPerRound()
        {
            return _changesPerRound;
        }

        public IList<int> GetCarrierChangesAfterEachRound()
        {
            return _changesAfterRound;
        }

        public double GetAverageCarrierChangesPerRound()
        {
            return _totalChanges / (double) _changesAfterRound.Count;
        }

        public int GetMaxChangesBetweenRounds()
        {
            return _changesAfterRound.Max();
        }

        public double GetCarrierChangeVariance()
        {
            double avg = GetAverageCarrierChangesPerRound();

            double sumSquaredError = 0;
            foreach (int changesAfterRound in _changesAfterRound) {
                sumSquaredError += Math.Pow(changesAfterRound - avg, 2.0);
            }

            return sumSquaredError / _changesAfterRound.Count;
        }
    }
}
