﻿using System;
using System.Collections.Generic;
using System.Linq;
using PaintShopSchedulingSimpleTypes;
using PaintShopSchedulingSimpleTypes.SimpleTypes;

namespace SimpleSolutionValidator.SolutionEvaluator
{
    public class ColorChangeEvaluator : IColorChangeEvaluator
    {
        private int _totalNumberOfColorChanges;
        private int _totalColorChangeCosts;

        private IList<int> _colorChangesPerRound;
        private IList<int> _colorCostsPerRound;
        private ICollection<SimpleColorTransitionCost> _colorTransitionCosts;

        public void Evaluate(ICollection<SimpleColorTransitionCost> colorTransitionCosts, ISimpleSolution solution)
        {
            _totalNumberOfColorChanges = 0;
            _totalColorChangeCosts = 0;
            _colorChangesPerRound = Enumerable.Repeat(0, solution.NumberOfRounds - 1).ToList();
            _colorCostsPerRound = Enumerable.Repeat(0, solution.NumberOfRounds - 1).ToList();
            _colorTransitionCosts = colorTransitionCosts;

            for (int roundIndex = 1; roundIndex < solution.NumberOfRounds; roundIndex++)
            {
                // check for color change to last carrier of previous round
                int numberOfCarriersInCurrentRound = solution.GetNumberOfCarriersInRound(roundIndex);
                if (numberOfCarriersInCurrentRound > 0)
                {
                    Position previousColoredPosition = GetPreviousColoredPosition(new Position(roundIndex, 0), solution);

                    if (previousColoredPosition.RoundIndex > -1)
                    {
                        int fromColorId =
                            solution.GetColorIdAtPosition(previousColoredPosition);
                        int toColorId = solution.GetColorIdAtPosition(new Position(roundIndex, 0));

                        UpdateColorChangeData(fromColorId, toColorId, roundIndex);
                    }
                }

                for (int posIndex = 0; posIndex < numberOfCarriersInCurrentRound - 1; posIndex++)
                {

                    int fromColorId =
                        solution.GetColorIdAtPosition(new Position(roundIndex, posIndex));
                    int toColorId = solution.GetColorIdAtPosition(new Position(roundIndex, posIndex + 1));

                    UpdateColorChangeData(fromColorId, toColorId, roundIndex);
                }
            }
        }

        private Position GetPreviousColoredPosition(Position position, ISimpleSolution solution)
        {
            Position previousPosition = PositionCalculator.GetPreviousPosition(position, solution);

            while (previousPosition.RoundIndex > -1)
            {
                if (solution.GetColorIdAtPosition(previousPosition) > -1)
                {
                    return previousPosition;
                }
                previousPosition = PositionCalculator.GetPreviousPosition(previousPosition, solution);
            }

            return new Position(-1, -1);
        }

        private void UpdateColorChangeData(int fromColorId, int toColorId, int roundIndex)
        {
            if (fromColorId != toColorId)
            {
                int transitionCosts = GetTransitionCosts(fromColorId, toColorId);
                _totalNumberOfColorChanges++;
                _totalColorChangeCosts += transitionCosts;

                _colorChangesPerRound[roundIndex - 1]++;
                _colorCostsPerRound[roundIndex - 1] += transitionCosts;
            }
        }

        private int GetTransitionCosts(int fromColorId, int toColorId)
        {
            int colorTransitionCosts = 0;

            foreach (SimpleColorTransitionCost transitionCost in _colorTransitionCosts)
            {
                if (transitionCost.FromColor.Id == fromColorId && transitionCost.ToColor.Id == toColorId)
                {
                    return transitionCost.Cost;
                }
            }

            return colorTransitionCosts;
        }

        public int GetTotalColorChanges()
        {
            return _totalNumberOfColorChanges;
        }

        public IList<int> GetColorChangesPerRound()
        {
            return _colorChangesPerRound;
        }

        public int GetTotalColorChangeCosts()
        {
            return _totalColorChangeCosts;
        }

        public IList<int> GetColorCostsPerRound()
        {
            return _colorCostsPerRound;
        }

        public long GetSquaredColorCosts()
        {
            long totalSquaredCosts = 0;

            foreach (int costPerRound in _colorCostsPerRound)
            {
                totalSquaredCosts += costPerRound * costPerRound;
            }

            return totalSquaredCosts;
        }

        public double GetAverageColorChangesPerRound()
        {
            return _totalNumberOfColorChanges / (double)_colorChangesPerRound.Count;
        }

        public int GetMaximumColorChangesInSingleRound()
        {
            return _colorChangesPerRound.Max();
        }

        public double GetColorChangeVariance()
        {
            double avg = GetAverageColorChangesPerRound();

            double sumSquaredError = 0;
            foreach (int changesInRound in _colorChangesPerRound) {
                sumSquaredError += Math.Pow(changesInRound - avg, 2.0);
            }

            return sumSquaredError / _colorChangesPerRound.Count;
        }
    }
}
