﻿using System.Collections.Generic;
using System.Linq;
using PaintShopSchedulingSimpleTypes;
using PaintShopSchedulingSimpleTypes.SimpleTypes;

namespace SimpleSolutionValidator.SolutionEvaluator
{
    public class MinBlockSizeEvaluator : IMinBlockSizeEvaluator
    {
        private int _totalViolations;
        private IList<int> _totalViolationsPerRound;
        private ICollection<SimpleCarrierTypeInformation> _carrierTypeInformations;
        private readonly SimpleInstance _instance;

        public MinBlockSizeEvaluator(SimpleInstance instance)
        {
            _instance = instance;
        }

        public void Evaluate(ICollection<SimpleCarrierTypeInformation> carrierTypeInformations, ISimpleSolution solution)
        {
            _totalViolations = 0;
            _totalViolationsPerRound = Enumerable.Repeat(0, solution.NumberOfRounds).ToList();
            _carrierTypeInformations = carrierTypeInformations;

            int currentBlockCarrierTypeId = -1;
            int currentBlockSize = 0;
            int blockEndRoundIndex = 0;
            int blockEndPositionIndex = 0;

            for (int roundIndex = 0; roundIndex < solution.NumberOfRounds; roundIndex++)
            {
                int numberOfCarriersInCurrentRound = solution.GetNumberOfCarriersInRound(roundIndex);

                for (int posIndex = 0; posIndex < numberOfCarriersInCurrentRound; posIndex++)
                {
                    int currentCarrierTypeId = solution.GetCarrierTypeIdAtPosition(new Position(roundIndex, posIndex));
                    if (currentBlockCarrierTypeId != currentCarrierTypeId)
                    {
                        if (blockEndRoundIndex > 0 || blockEndPositionIndex == solution.GetNumberOfCarriersInRound(0) - 1)
                        {
                            UpdateMinBlockViolations(currentBlockCarrierTypeId, currentBlockSize, blockEndRoundIndex, roundIndex, posIndex);
                        }

                        currentBlockCarrierTypeId = currentCarrierTypeId;
                        currentBlockSize = 1;
                    }
                    else
                    {
                        currentBlockSize++;
                    }
                    blockEndRoundIndex = roundIndex;
                    blockEndPositionIndex = posIndex;
                }
            }
            UpdateMinBlockViolations(currentBlockCarrierTypeId, currentBlockSize, blockEndRoundIndex, solution.NumberOfRounds-1, solution.GetNumberOfCarriersInRound(solution.NumberOfRounds-1));
        }

        private void UpdateMinBlockViolations(int currentBlockCarrierTypeId, int currentBlockSize, int blockEndRoundIndex, int roundIndex, int posIndex)
        {
            int minBlockSize =
                (from carrierTypeInfo in _carrierTypeInformations
                 where carrierTypeInfo.CarrierType.Id == currentBlockCarrierTypeId
                 select carrierTypeInfo.MinimumBlockSize).FirstOrDefault();

            if (currentBlockSize < minBlockSize)
            {
                _totalViolations++;
                _totalViolationsPerRound[blockEndRoundIndex]++;

                System.Console.WriteLine("The minimum block length of carrier type " +
                                         _instance.CarrierTypeNames[currentBlockCarrierTypeId] +
                                         " is violated in round " + roundIndex + " at position " + posIndex +
                                         ". The minimum block length for this type is " + minBlockSize +
                                         ", but the block length is " + currentBlockSize);
            }
        }

        public int GetTotalViolations()
        {
            return _totalViolations;
        }

        public IList<int> GetViolationsPerRound()
        {
            return _totalViolationsPerRound;
        }
    }
}