﻿using System.Collections.Generic;
using System.IO;
using PaintShopSchedulingSimpleTypes;

namespace SimpleSolutionValidator.SolutionEvaluator
{
    public class SimpleSolutionEvaluator : ISimpleSolutionEvaluator
    {
        // These constants can be configured for the console display for the round/violation matrix
        private const int TEXT_CELL_LENGTH = 5;
        private const int FIRST_TEXT_CELL_LENGTH = 30;

        public void EvaluateSolution(SimpleInstance instance, ISimpleSolution solution, TextWriter textWriter)
        {
            // Get number of carriers
            INumberOfCarriersEvaluator numberOfCarriersEvaluator = new NumberOfCarriersEvaluator();
            numberOfCarriersEvaluator.Evaluate(solution);
            int totalNumberOfCarriers = numberOfCarriersEvaluator.GetTotalNumberOfCarriers();
            IList<int> numberOfCarriersPerRound = numberOfCarriersEvaluator.GetNumberOfCarriersPerRound();

            // 
            // Soft constraints
            //
            ICarrierChangeEvaluator carrierChangeEvaluator = new CarrierChangeEvaluator();
            carrierChangeEvaluator.Evaluate(solution);
            int totalNumberOfCarrierChanges = carrierChangeEvaluator.GetTotalNumberOfChanges();
            int squaredNumberOfCarrierChanges = carrierChangeEvaluator.GetSquaredNumberOfChanges();
            IList<int> numberOfCarrierChangesPerRound = carrierChangeEvaluator.GetCarrierChangesPerRound();
            IList<int> numberOfCarrierChangesAfterEachRound = carrierChangeEvaluator.GetCarrierChangesAfterEachRound();

            IColorChangeEvaluator colorChangeEvaluator = new ColorChangeEvaluator();
            colorChangeEvaluator.Evaluate(instance.ColorTransitionCosts, solution);
            int totalColorChanges = colorChangeEvaluator.GetTotalColorChanges();
            IList<int> colorChangesPerRound = colorChangeEvaluator.GetColorChangesPerRound();
            int totalColorCosts = colorChangeEvaluator.GetTotalColorChangeCosts();
            long squaredColorCosts = colorChangeEvaluator.GetSquaredColorCosts();
            IList<int> colorCostsPerRound = colorChangeEvaluator.GetColorCostsPerRound();

            //
            // Hard constraints
            //

            // Unfulfilled demand constraint
            IUnfulfilledDemandEvaluator demandEvaluator = new UnfulfilledDemandEvaluator();
            demandEvaluator.Evaluate(instance.Demands, instance.Configurations, solution);
            int totalUnfulfilledDemands = demandEvaluator.GetUnfulfilledDemands();
            int totalMissingDemandPieces = demandEvaluator.GetMissingDemandPieces();
            IList<int> unfulfilledDemandsPerRound = demandEvaluator.GetUnfulfilledDemandsPerRound();
            IList<int> missingDemandPiecesPerRound = demandEvaluator.GetMissingDemandPiecesPerRound();

            // Min Block constraint
            IMinBlockSizeEvaluator minBlockEvaluator = new MinBlockSizeEvaluator(instance);
            minBlockEvaluator.Evaluate(instance.CarrierTypeInfos, solution);
            int totalMinBlockViolations = minBlockEvaluator.GetTotalViolations();
            IList<int> minBlockViolationsPerRound = minBlockEvaluator.GetViolationsPerRound();

            // Max Block constraint
            IMaxBlockSizeEvaluator maxBlockEvaluator = new MaxBlockSizeEvaluator(instance);
            maxBlockEvaluator.Evaluate(instance.CarrierTypeInfos, solution);
            int totalMaxBlockViolations = maxBlockEvaluator.GetTotalViolations();
            IList<int> maxBlockViolationsPerRound = maxBlockEvaluator.GetViolationsPerRound();

            // Round Capacity constraint
            IRoundCapacityEvaluator roundCapacityEvaluator = new RoundCapacityEvaluator();
            roundCapacityEvaluator.Evaluate(instance.MinCapacity, instance.MaxCapacity, solution);
            int totalCapacityViolations = roundCapacityEvaluator.GetTotalViolations();
            IList<int> capacityViolationsPerRound = roundCapacityEvaluator.GetViolationsPerRound();

            // Carrier Availability constraint
            ICarrierAvailabilityEvaluator carrierAvailabilityEvaluator = new CarrierAvailabilityEvaluator(instance);
            carrierAvailabilityEvaluator.Evaluate(instance.CarrierTypeInfos, solution);
            int totalAvailabilityViolations = carrierAvailabilityEvaluator.GetTotalViolations();
            IList<int> availabilityViolationsPerRound = carrierAvailabilityEvaluator.GetViolationsPerRound();

            // Forbidden Carrier Type Sequences constraint
            IForbiddenCarrierTypeSequencesEvaluator forbiddenCarrierTypeSequencesEvaluator = new ForbiddenCarrierTypeSequencesEvaluator(instance);
            forbiddenCarrierTypeSequencesEvaluator.Evaluate(instance.ForbiddenCarrierSequences, solution);
            int totalForbiddenCarrierTypeSequenceViolations = forbiddenCarrierTypeSequencesEvaluator.GetTotalViolations();
            IList<int> forbiddenCarrierTypeSequencesPerRound = forbiddenCarrierTypeSequencesEvaluator.GetViolationsPerRound();

            // Forbidden Color Sequences constraint
            IForbiddenColorSequencesEvaluator forbiddenColorSequencesEvaluator = new ForbiddenColorSequencesEvaluator(instance);
            forbiddenColorSequencesEvaluator.Evaluate(instance.ForbiddenColorSequences, solution);
            int totalForbiddenColorSequenceViolations = forbiddenColorSequencesEvaluator.GetTotalViolations();
            IList<int> forbiddenColorSequencesPerRound = forbiddenColorSequencesEvaluator.GetViolationsPerRound();

            // Print feasability and cost
            int totalHardConstraintViolations = totalMissingDemandPieces +
                                                totalMaxBlockViolations + totalCapacityViolations +
                                                totalAvailabilityViolations + totalForbiddenCarrierTypeSequenceViolations +
                                                totalForbiddenColorSequenceViolations;
            long totalSoftConstraintViolations = squaredNumberOfCarrierChanges + squaredColorCosts;
            if (totalHardConstraintViolations == 0) {
                textWriter.WriteLine("FEASIBLE");
            }
            else {
                textWriter.WriteLine("INFEASIBLE");
                textWriter.WriteLine("Total number of violations: " + totalHardConstraintViolations);
            }

            textWriter.WriteLine("Total solution cost: " + totalSoftConstraintViolations);

            textWriter.WriteLine();
            textWriter.WriteLine();

            //
            // Print general output
            //
            textWriter.WriteLine("Total number of carriers: " + totalNumberOfCarriers);

            PrintShortHorizontalLine('=', textWriter);
            textWriter.WriteLine("Soft constraints");
            PrintShortHorizontalLine('=', textWriter);
            textWriter.WriteLine();

            textWriter.WriteLine("Color Changes: " + totalColorChanges);
            textWriter.WriteLine("Color Costs: " + totalColorCosts);
            textWriter.WriteLine("Squared color costs: " + squaredColorCosts);

            textWriter.WriteLine("Carrier Changes: " + totalNumberOfCarrierChanges);
            textWriter.WriteLine("Squared Carrier Changes: " + squaredNumberOfCarrierChanges);
            //textWriter.WriteLine("Carrier Changes (heuristic): " + heuristicNumberOfCarrierChanges);
            //textWriter.WriteLine("Squared Carrier Changes (heuristic): " + squaredHeuristicNumberOfCarrierChanges);

            textWriter.WriteLine("Unfulfilled Demands: " + totalUnfulfilledDemands);
            textWriter.WriteLine("Missing Demand Pieces: " + totalMissingDemandPieces);

            PrintShortHorizontalLine('=', textWriter);
            textWriter.WriteLine("Hard constraints");
            PrintShortHorizontalLine('=', textWriter);
            textWriter.WriteLine();

            textWriter.WriteLine("Min Block Size Violations: " + totalMinBlockViolations);
            textWriter.WriteLine("Max Block Size Violations: " + totalMaxBlockViolations);

            textWriter.WriteLine("Round Capacity Violations: " + totalCapacityViolations);

            textWriter.WriteLine("Carrier Availability Violations: " + totalAvailabilityViolations);
            textWriter.WriteLine("Forbidden Carrier Sequence Violations: " + totalForbiddenCarrierTypeSequenceViolations);
            textWriter.WriteLine("Forbidden Color Sequence Violations: " + totalForbiddenColorSequenceViolations);

            PrintShortHorizontalLine('-', textWriter);

            //
            // Print violations per round
            //
            PrintHorizontalLine(solution.NumberOfRounds, '=', textWriter);
            PrintRoundHeaders(solution.NumberOfRounds, textWriter);

            // Information
            PrintHorizontalLine(solution.NumberOfRounds, '-', textWriter);
            PrintFirstCellText("Number of carriers", textWriter);
            PrintIntegerValuesPerRound(numberOfCarriersPerRound, false, false, textWriter);

            // Soft constraints
            PrintHorizontalLine(solution.NumberOfRounds, '=', textWriter);
            PrintFirstCellText("Carrier changes", textWriter);
            PrintIntegerValuesPerRound(numberOfCarrierChangesPerRound, false, false, textWriter);

            PrintHorizontalLine(solution.NumberOfRounds, '=', textWriter);
            PrintFirstCellText("Carrier changes after round", textWriter);
            PrintIntegerValuesPerRound(numberOfCarrierChangesAfterEachRound, false, true, textWriter);

            //PrintHorizontalLine(solution.NumberOfRounds, '=', textWriter);
            //PrintFirstCellText("Carrier changes (heuristic)", textWriter);
            //PrintIntegerValuesPerRound(heuristicNumberOfCarrierChangesPerRound, false, false, textWriter);

            //PrintHorizontalLine(solution.NumberOfRounds, '=', textWriter);
            //PrintFirstCellText("Heuristic changes after round", textWriter);
            //PrintIntegerValuesPerRound(heuristicNumberOfCarrierChangesAfterEachRound, false, true, textWriter);

            PrintHorizontalLine(solution.NumberOfRounds, '-', textWriter);
            PrintFirstCellText("Color changes", textWriter);
            PrintIntegerValuesPerRound(colorChangesPerRound, true, false, textWriter);

            PrintHorizontalLine(solution.NumberOfRounds, '-', textWriter);
            PrintFirstCellText("Color change costs", textWriter);
            PrintIntegerValuesPerRound(colorCostsPerRound, true, false, textWriter);

            PrintHorizontalLine(solution.NumberOfRounds, '-', textWriter);
            PrintFirstCellText("Unfulfilled demands", textWriter);
            PrintIntegerValuesPerRound(unfulfilledDemandsPerRound, false, false, textWriter);

            PrintHorizontalLine(solution.NumberOfRounds, '-', textWriter);
            PrintFirstCellText("Missing pieces", textWriter);
            PrintIntegerValuesPerRound(missingDemandPiecesPerRound, false, false, textWriter);

            // Hard constraints
            PrintHorizontalLine(solution.NumberOfRounds, '-', textWriter);
            PrintFirstCellText("Min Block", textWriter);
            PrintIntegerValuesPerRound(minBlockViolationsPerRound, false, false, textWriter);

            PrintHorizontalLine(solution.NumberOfRounds, '-', textWriter);
            PrintFirstCellText("Max Block", textWriter);
            PrintIntegerValuesPerRound(maxBlockViolationsPerRound, true, false, textWriter);

            PrintHorizontalLine(solution.NumberOfRounds, '-', textWriter);
            PrintFirstCellText("Carrier availability", textWriter);
            PrintIntegerValuesPerRound(availabilityViolationsPerRound, true, false, textWriter);

            PrintHorizontalLine(solution.NumberOfRounds, '-', textWriter);
            PrintFirstCellText("Round capacity", textWriter);
            PrintIntegerValuesPerRound(capacityViolationsPerRound, true, false, textWriter);

            PrintHorizontalLine(solution.NumberOfRounds, '-', textWriter);
            PrintFirstCellText("Forbidden CarrSeq", textWriter);
            PrintIntegerValuesPerRound(forbiddenCarrierTypeSequencesPerRound, true, false, textWriter);

            PrintHorizontalLine(solution.NumberOfRounds, '-', textWriter);
            PrintFirstCellText("Forbidden ColorSeq", textWriter);
            PrintIntegerValuesPerRound(forbiddenColorSequencesPerRound, true, false, textWriter);

            // Print Footer
            PrintHorizontalLine(solution.NumberOfRounds, '=', textWriter);
        }

        public void EvaluateSolutionSingleLineCost(string instanceName, SimpleInstance instance, ISimpleSolution solution, TextWriter textWriter)
        {
            // 
            // Soft constraints
            //
            ICarrierChangeEvaluator carrierChangeEvaluator = new CarrierChangeEvaluator();
            carrierChangeEvaluator.Evaluate(solution);
            int squaredNumberOfCarrierChanges = carrierChangeEvaluator.GetSquaredNumberOfChanges();
            double avgCarrierChangesPerRound = carrierChangeEvaluator.GetAverageCarrierChangesPerRound();
            int maxCarrierChangesPerRound = carrierChangeEvaluator.GetMaxChangesBetweenRounds();
            double carrierChangeVariance = carrierChangeEvaluator.GetCarrierChangeVariance();

            IColorChangeEvaluator colorChangeEvaluator = new ColorChangeEvaluator();
            colorChangeEvaluator.Evaluate(instance.ColorTransitionCosts, solution);
            long squaredColorCosts = colorChangeEvaluator.GetSquaredColorCosts();
            double avgColorChangesPerRound = colorChangeEvaluator.GetAverageColorChangesPerRound();
            int maxColorChangesPerRound = colorChangeEvaluator.GetMaximumColorChangesInSingleRound();
            double colorChangeVariance = colorChangeEvaluator.GetColorChangeVariance();

            //
            // Hard constraints
            //

            // Unfulfilled demand constraint
            IUnfulfilledDemandEvaluator demandEvaluator = new UnfulfilledDemandEvaluator();
            demandEvaluator.Evaluate(instance.Demands, instance.Configurations, solution);
            int totalMissingDemandPieces = demandEvaluator.GetMissingDemandPieces();

            // Min Block constraint
            IMinBlockSizeEvaluator minBlockEvaluator = new MinBlockSizeEvaluator(instance);
            minBlockEvaluator.Evaluate(instance.CarrierTypeInfos, solution);
            int totalMinBlockViolations = minBlockEvaluator.GetTotalViolations();

            // Max Block constraint
            IMaxBlockSizeEvaluator maxBlockEvaluator = new MaxBlockSizeEvaluator(instance);
            maxBlockEvaluator.Evaluate(instance.CarrierTypeInfos, solution);
            int totalMaxBlockViolations = maxBlockEvaluator.GetTotalViolations();

            // Round Capacity constraint
            IRoundCapacityEvaluator roundCapacityEvaluator = new RoundCapacityEvaluator();
            roundCapacityEvaluator.Evaluate(instance.MinCapacity, instance.MaxCapacity, solution);
            int totalCapacityViolations = roundCapacityEvaluator.GetTotalViolations();

            // Carrier Availability constraint
            ICarrierAvailabilityEvaluator carrierAvailabilityEvaluator = new CarrierAvailabilityEvaluator(instance);
            carrierAvailabilityEvaluator.Evaluate(instance.CarrierTypeInfos, solution);
            int totalAvailabilityViolations = carrierAvailabilityEvaluator.GetTotalViolations();

            // Forbidden Carrier Type Sequences constraint
            IForbiddenCarrierTypeSequencesEvaluator forbiddenCarrierTypeSequencesEvaluator = new ForbiddenCarrierTypeSequencesEvaluator(instance);
            forbiddenCarrierTypeSequencesEvaluator.Evaluate(instance.ForbiddenCarrierSequences, solution);
            int totalForbiddenCarrierTypeSequenceViolations = forbiddenCarrierTypeSequencesEvaluator.GetTotalViolations();

            // Forbidden Color Sequences constraint
            IForbiddenColorSequencesEvaluator forbiddenColorSequencesEvaluator = new ForbiddenColorSequencesEvaluator(instance);
            forbiddenColorSequencesEvaluator.Evaluate(instance.ForbiddenColorSequences, solution);
            int totalForbiddenColorSequenceViolations = forbiddenColorSequencesEvaluator.GetTotalViolations();

            int totalHardConstraintViolations = totalMissingDemandPieces +
                                                totalMaxBlockViolations + totalCapacityViolations +
                                                totalAvailabilityViolations + totalForbiddenCarrierTypeSequenceViolations +
                                                totalForbiddenColorSequenceViolations;
            long totalSoftConstraintViolations = squaredNumberOfCarrierChanges + squaredColorCosts;

            string csvLine = instanceName + ";" +
                              totalHardConstraintViolations + ";" +
                              totalMinBlockViolations + ";" +
                              (totalHardConstraintViolations + totalMinBlockViolations) + ";";

            if (totalHardConstraintViolations + totalMinBlockViolations == 0)
            {
                // we only include cost information if the solution is feasible
                csvLine += totalSoftConstraintViolations + ";" +
                            avgColorChangesPerRound + ";" +
                            avgCarrierChangesPerRound + ";" +
                            maxColorChangesPerRound + ";" +
                            maxCarrierChangesPerRound + ";" +
                            colorChangeVariance + ";" +
                            carrierChangeVariance;
            }
            else {
                csvLine += ";;;;;;";
            }


            textWriter.WriteLine(csvLine);
        }

        private static void PrintRoundHeaders(int numberOfRounds, TextWriter textWriter)
        {
            // First cell of header is empty
            textWriter.Write(new string(' ', FIRST_TEXT_CELL_LENGTH));
            textWriter.Write("|");

            string cellLengthFormat = "{0,-" + (TEXT_CELL_LENGTH - 1) + "}";

            for (int roundIndex = 0; roundIndex < numberOfRounds; roundIndex++)
            {
                textWriter.Write(" R");

                string numString = string.Format(cellLengthFormat, roundIndex);
                textWriter.Write(numString.Substring(0, TEXT_CELL_LENGTH - 1));

                textWriter.Write(" |");
            }

            textWriter.WriteLine();
        }

        private static void PrintHorizontalLine(int numberOfRounds, char lineChar, TextWriter textWriter)
        {
            textWriter.WriteLine(new string(lineChar,
                FIRST_TEXT_CELL_LENGTH + (TEXT_CELL_LENGTH + 3) * numberOfRounds));
        }

        private static void PrintShortHorizontalLine(char lineChar, TextWriter textWriter)
        {
            textWriter.WriteLine(new string(lineChar, 20));
        }

        private static void PrintIntegerValuesPerRound(IList<int> values, bool blankHistoryRound, bool blankLastRound, TextWriter textWriter)
        {
            textWriter.Write("|");

            string cellLengthFormat = "{0," + TEXT_CELL_LENGTH + "}";

            if (blankHistoryRound)
            {
                textWriter.Write(" " + new string(' ', TEXT_CELL_LENGTH) + " |");
            }

            foreach (int value in values)
            {
                textWriter.Write(" ");

                string numString = string.Format(cellLengthFormat, value);
                textWriter.Write(numString.Substring(0, TEXT_CELL_LENGTH));

                textWriter.Write(" |");
            }

            if (blankLastRound)
            {
                textWriter.Write(" " + new string(' ', TEXT_CELL_LENGTH) + " |");
            }

            textWriter.WriteLine();
        }

        private static void PrintFirstCellText(string text, TextWriter textWriter)
        {
            string firstCellLengthFormat = "{0," + FIRST_TEXT_CELL_LENGTH + "}";
            string formattedString = string.Format(firstCellLengthFormat, text);

            textWriter.Write(formattedString.Substring(0, FIRST_TEXT_CELL_LENGTH));
        }
    }
}
